#include "cas_codec.h"

//    This file is part of CAS.

//    CAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    CAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

QByteArray wrNum(uchar numOfBits, quint64 value, bool be)
{
    QByteArray ret(numOfBits / 8, (char) 0);

    if (be) value = qToBigEndian(value);
    else    value = qToLittleEndian(value);

    memcpy(ret.data(), &value, ret.size());

    return ret;
}

quint64 rdNum(const QByteArray &dataIn, bool be)
{
    quint64 ret = 0;

    memcpy(&ret, dataIn.data(), dataIn.size());

    if (be) return qFromBigEndian(ret);
    else    return qFromLittleEndian(ret);
}

quint64 rdNum(QIODevice *dev, uchar numOfBits, bool be)
{
    return rdNum(dev->read(numOfBits / 8), be);
}

Channel::Channel(uchar chId, uchar *bits, bool *be, QObject *parent) : QObject(parent)
{
    channelId       = chId;
    pcmBits         = bits;
    bigE            = be;
    pcmSample       = 0;
    xGap            = 0;
    firstReadSample = true;
    ramping         = false;
}

void Channel::toCas(uchar id, const QByteArray &pcm, bool forceWr)
{
    if (id == channelId)
    {
        quint64 currentSample = rdNum(pcm, *bigE);
        bool    wrCasSample   = false;

        if (firstReadSample)
        {
            // at the very begining of the PCM stream, the codec does
            // not know if the signal is ramping or not and xGap is
            // assumed to be 1 so the first PCM sample is always
            // written to set a base line.

            xGap            = 1;
            wrCasSample     = true;
            firstReadSample = false;
        }
        else if (ramping)
        {
            if (currentSample < pcmSample)
            {
                // the currently read pcm sample that reads
                // a lower value when the loop is expecting
                // a higher value is considered a peak of a
                // sinewave signal so a CAS sample needs to
                // be written at this point.

                // doing this captures the upper peak of the
                // sinewave.

                ramping     = false;
                wrCasSample = true;
            }
        }
        else
        {
            if (currentSample > pcmSample)
            {
                // this does the exact opposite of the above.
                // it captures the lower peak of the sinewave.

                ramping     = true;
                wrCasSample = true;
            }
        }

        if (wrCasSample || forceWr || (xGap == 255))
        {
            // CAS format: [xGap][PCM_Sample]

            emit sampleRdy(channelId, wrNum(8, xGap, *bigE) + wrNum(*pcmBits, pcmSample, *bigE));

            xGap = 0;
        }

        xGap++;

        pcmSample = currentSample;
    }
}

void Channel::fromCas(uchar id, const QByteArray &cas)
{
    if (id == channelId)
    {
        xGap = rdNum(cas.left(1), *bigE);

        quint64 currentSample = rdNum(cas.mid(1), *bigE);
        quint64 prevSample    = pcmSample;
        quint64 yGap          = 0;

        if (currentSample > prevSample)
        {
            yGap = (currentSample - prevSample) / xGap;
        }
        else if (currentSample < prevSample)
        {
            yGap = (prevSample - currentSample) / xGap;
        }

        for (; xGap != 0; --xGap)
        {
            if (currentSample > prevSample)
            {
                // a current sample higher than previously read sample is
                // assumed to be a ramping up sinewave so it increments
                // the output pcm samples accordingly.

                pcmSample += yGap;
            }
            else if (currentSample < prevSample)
            {
                // this is the opposite of the above.

                pcmSample -= yGap;
            }

            emit sampleRdy(channelId, wrNum(*pcmBits, pcmSample, *bigE));
        }
    }
}

CASCodec::CASCodec(QObject *parent) : QObject(0)
{
    QThread *thread = new QThread();

    connect(this, SIGNAL(destroyed()), thread, SLOT(quit()));
    connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));

    moveToThread(thread);

    if (parent)
    {
        connect(parent, SIGNAL(destroyed()), this, SLOT(deleteLater()));
    }

    currentChannel = 0;
    sampleRate     = 0;
    casRev         = 0;
    srcIO          = 0;
    dstIO          = 0;

    stop();

    thread->start();
}

bool CASCodec::rdCasHeader(QIODevice *rdIO)
{
    bool       ret    = false;
    QByteArray header = rdIO->read(10);

    if (header.startsWith(CAS_TAG) && (header.size() == 10))
    {
        casRev     = header[3];
        pcmBits    = header[4];
        sampleRate = rdNum(header.mid(6, 4), false);
        ret        = true;

        emit updatePCMBits(pcmBits);
        emit updateNumOfChannels(header[5]);
        emit updateSampleRate(sampleRate);
    }

    return ret;
}

bool CASCodec::wrCasHeader(QIODevice *wrIO)
{
    bool       ret = false;
    QByteArray header;

    casRev = CAS_REV;

    header.append(CAS_TAG);
    header.append(CAS_REV);
    header.append(pcmBits);
    header.append((uchar) buffer.size());
    header.append(wrNum(32, sampleRate, false));

    if (wrIO->write(header) != -1)
    {
        ret = true;
    }

    return ret;
}

void CASCodec::setEncodeMode(bool state)    {encode     = state;}
void CASCodec::setBigEndianMode(bool state) {bigE       = state;}
void CASCodec::setSampleRate(int rate)      {sampleRate = rate;}
void CASCodec::setPCMBits(int numOfBits)    {pcmBits    = numOfBits;}

void CASCodec::setNumOfChannels(int ch)
{
    emit delChannels();

    buffer.clear();

    for (int i = 0; i < ch; ++i)
    {
        Channel *channel = new Channel(i, &pcmBits, &bigE);

        connect(this, SIGNAL(delChannels()), channel, SLOT(deleteLater()));
        connect(this, SIGNAL(encodeSample(uchar,QByteArray,bool)), channel, SLOT(toCas(uchar,QByteArray,bool)));
        connect(this, SIGNAL(decodeSample(uchar,QByteArray)), channel, SLOT(fromCas(uchar,QByteArray)));
        connect(channel, SIGNAL(sampleRdy(uchar,QByteArray)), this, SLOT(finishedSample(uchar,QByteArray)));

        buffer.append(QList<QByteArray>());
    }
}

void CASCodec::dataIn()
{
    if (srcIO && dstIO)
    {
        while (!stopRequested && srcIO->bytesAvailable())
        {
            bool forceWr = false;

            for (int i = 0; i < buffer.size(); ++i)
            {
                if (encode)
                {
                    if (!srcIO->isSequential())
                    {
                        forceWr = (lastBytes >= srcIO->bytesAvailable());
                    }

                    QByteArray data = srcIO->read(pcmBits / 8);

                    emit encodeSample(i, data, forceWr);
                }
                else
                {
                    QByteArray data = srcIO->read((pcmBits / 8) + 1);

                    emit decodeSample(i, data);
                }
            }
        }
    }
}

void CASCodec::finishedSample(uchar id, const QByteArray &data)
{
    buffer[id].append(data);

    bool allChannelsPending = true;

    for (int i = 0; i < buffer.size(); ++i)
    {
        if (buffer[i].isEmpty())
        {
            allChannelsPending = false;

            break;
        }
    }

    if (allChannelsPending)
    {
        for (int i = 0; i < buffer.size(); ++i)
        {
            dstIO->write(buffer[i].takeFirst());
        }
    }
}

void CASCodec::exec(QIODevice *rdIO, QIODevice *wrIO)
{
    if (buffer.size() == 0)
    {
        emit stopError(tr("The number of channels cannot be zero."));
    }
    else if (!rdIO->isOpen() || !rdIO->isReadable())
    {
        emit stopError(tr("Input io device is not open or readable."));
    }
    else if (!wrIO->isOpen() || !wrIO->isWritable())
    {
        emit stopError(tr("Output io device is not open or writable."));
    }
    else
    {
        stopRequested = false;
        lastBytes     = (pcmBits / 8) * buffer.size();
        srcIO         = rdIO;
        dstIO         = wrIO;

        emit runningState(true);

        if (encode) emit codecStatus(tr("Encoding."));
        else        emit codecStatus(tr("Decoding."));

        connect(rdIO, SIGNAL(readyRead()), this, SLOT(dataIn()));

        if (!rdIO->isSequential())
        {
            emit rdIO->readyRead();
        }
    }
}

void CASCodec::stop()
{
    if (srcIO)
    {
        disconnect(srcIO, SIGNAL(readyRead()), this, SLOT(dataIn()));
    }

    setNumOfChannels(buffer.size());

    stopRequested  = true;
    currentChannel = 0;
    lastBytes      = 0;
    srcIO          = 0;
    dstIO          = 0;

    emit runningState(false);
    emit codecStatus(tr("Ready."));
}
